@@ -2,9 +2,12 @@ from typing import Optional
 
 
 class Group:
-    group_id: Optional[str]
     members: list[str]
+    group_id: Optional[str]
 
-    def __init__(self, members: list[str], group_id: Optional[str]=None):
+    def __init__(self,
+        members: list[str],
+        group_id: Optional[str]=None,
+    ):
         self.members = members
         self.group_id = group_id
